#!/usr/bin/env python
# -*- coding: utf-8 -*-
# =============================================================================
## @file ostap/fitting/convolution.py
#  Set of useful basic utilities to use convolution 
#  @author Vanya BELYAEV Ivan.Belyaeve@itep.ru
#  @date 2011-07-25
# =============================================================================
"""Set of useful basic utilities to use convoltuion"""
# =============================================================================
__version__ = "$Revision:"
__author__  = "Vanya BELYAEV Ivan.Belyaev@itep.ru"
__date__    = "2011-07-25"
__all__     = (
    ##
    'ConvolutionConfig', ## context manager  to provide convolution parmaeters 
    'Convolution'      , ## helper utility to build convolution
    'Convolution_pdf'  , ## ``ready-to-use'' PDF for convolution 
    )
# =============================================================================
from   ostap.fitting.pdfbasic import PDF1, Generic1D_pdf
from   ostap.core.ostap_types import num_types ,  integer_types 
import ROOT, math
# =============================================================================
from   ostap.logger.logger import getLogger
if '__main__' ==  __name__ : logger = getLogger ( 'ostap.fitting.convolution' )
else                       : logger = getLogger ( __name__         )
# =============================================================================
## helper class to keep convoltuon configuration
class CnvConfig(object) :
    """helper class to keep convoltuon configuration"""
    _config = {}
    ## retreive (copy) of current configuration
    @classmethod 
    def config ( klass ) :
        """Return copy of currrent configuration
        """
        return klass._config.copy()
    
# ============================================================================
## Context manager to treat the convolutoi nconfiguration
class ConvolutionConfig(object):
    def __init__ ( self , **confif ) :
        self.__config = config
    ## context manager 
    def __enter__ ( self ) :
        self.__keep     = _CnvCnf._config.copy()
        _CnvCnf._config =  self.__config.copy()
        return self
    ## context manager 
    def __exit__ ( self , *_ ) :
        ## resore previous configruation 
        _CnvCnf._config =  self.__keep.copy()
        
# =============================================================================
## @class Convolution
#  Helper class to perform the convolution
#  @code
#  >>> pdf  = .... ##  original PDF (the one from Ostap or bare ROOT.RooAbsPdf) 
#  >>> resolution = 3 * MeV                                     ## fixed number ``sigma''
#  >>> # resolution = ROOT.RooRealVar( 'sigma','',2*MeV,4*MeV)  ## ``sigma'' as ROOT.RooAbsReal 
#  >>> # resolution = ResoGauss_pdf ( ... )                     ## pdf from Ostap
#  >>> # resolution = ...                                       ## bare ROOT.RooAbsPdf
#  >>> cnv = Convolution ('CNV' , pdf , xvar  =  xvar , resolution = resolution )
#  >>> cnv_pdf = cnv.pdf 
#  @endcode
#  @author Vanya BELYAEV Ivan.Belyaev@itep.ru
#  @date 2014-07-13
class Convolution(object):
    """Helper class to make a convolution PDF:
    
    >>> pdf  = .... ##  original PDF (the one from Ostap or bare ROOT.RooAbsPdf) 
    >>> resolution = 3 * MeV                                     ## fixed number ``sigma''
    >>> # resolution = ROOT.RooRealVar( 'sigma','',2*MeV,4*MeV)  ## ``sigma'' as ROOT.RooAbsReal 
    >>> # resolution = ResoGauss_pdf ( ... )                     ## pdf from Ostap
    >>> # resolution = ...                                       ## bare ROOT.RooAbsPdf
    >>> cnv = Convolution ('CNV' , pdf , xvar  =  xvar , resolution = resolution )
    >>> cnv_pdf = cnv.pdf 
    """
    def __init__ ( self              ,
                   pdf               ,   ## the PDF to be convoluted 
                   xvar              ,   ## the axis variable
                   resolution        ,   ## the resolution
                   name     = ''     ,   ## name for Gaussian resolution pdf 
                   useFFT   = True   ,   ## use FFT ? 
                   nbins    = 10000  ,   ## number of bins for FFT
                   buffer   = 0.25   ,   ## buffer fraction use for setBufferFraction
                   bufstrat = None   ,   ## "Buffer strategy" : (0,1,2)
                   shift1   = None   ,   ## shift1 parameter
                   shift2   = None   ,   ## shift2 parameter
                   nsigmas  = 6      ) : ## number of sigmas for setConvolutionWindow

        ## the axis 
        assert isinstance ( xvar , ROOT.RooAbsReal ) , "``xvar'' must be ROOT.RooAbsReal"
        
        self.__xvar   = xvar
        self.__useFFT = True if useFFT else False

        self.__arg_pdf        = pdf
        self.__arg_resolution = resolution
        self.__arg_xvar       = xvar
        
        ## pdf itself 
        if   isinstance ( pdf , PDF1           ) : self.__old_pdf = pdf
        elif isinstance ( pdf , ROOT.RooAbsPdf ) :
            self.__old_pdf = Generic1D_pdf ( pdf , xvar = self.__xvar )
        else :
            raise TypeError("Convolution: invalid ``pdf'' %s/%s"  % ( pdf , type ( pdf ) ) )
        
        ## resolution  function 
        if   isinstance ( resolution , PDF1           ) :
            self.__resolution = resolution
        elif isinstance ( resolution , ROOT.RooAbsPdf ) :
            self.__resolution = Generic1D_pdf ( resolution , xvar = self.__xvar ) 
        else :
            ## use   Gaussian resolution
            import ostap.fitting.resolution as OFR
            rname  = name if name else 'ResoGauss'
            rname  = self.old_pdf.generate_name ( prefix = rname )            
            self.__resolution = OFR.ResoGauss ( rname              ,
                                                self.__xvar        ,
                                                sigma = resolution )
        self.__nbins    = nbins
        self.__buffer   = buffer
        self.__bufstrat = bufstrat 
        self.__nsigmas  = nsigmas
        self.__shift1   = shift1 
        self.__shift2   = shift2

        name = name if name else PDF1.generate_name ( prefix = 'cnv_%s@%s' % ( pdf.name , self.resolution.name ) )
        
        if self.useFFT : ## Use Fast Fourier transform  (fast)
            
            assert isinstance ( nbins  , integer_types ) and 500   <  abs ( nbins  )  , \
                   "Invalid ``nbins''  parameter %s/%s for fast Fourier transform"  % ( nbins  , type ( nbins  ) )
            assert isinstance ( buffer ,  float        ) and 0.03  <= buffer <=0.9    , \
                   "Invalid ``buffer'' parameter %s/%s for ``setBufferFraction''"   % ( buffer , type ( buffer ) )

            ## adjust #bins if positive. keep it as it is if negavtive 
            if hasattr ( self.__resolution , 'sigma' ) and self.__xvar.minmax() and  self.__nbins > 0 :
                mn , mx = self.xvar.minmax()
                dm  = mx - mn
                sv  = self.__resolution.sigma.getVal() 
                dm /= sv
                nb  = min ( 50 * ( int ( dm ) + 1  ) , 2**14 )
                nb  = 2**math.frexp(nb)[1]
                if nb > self.nbinsFFT : 
                    self.__nbins  = nb       
                    logger.info('Convolution: choose #bins %d' % self.__nbins )

            self.__xvar.setBins ( self.nbinsFFT , 'cache' )

            self.__pdf = ROOT.RooFFTConvPdf (
                self.old_pdf.new_roo_name ( 'fft' ) ,
                'FFT convolution: %s (*) %s' %  ( pdf.name , self.resolution.name ) ,
                self.__xvar              ,
                self.__old_pdf    .pdf   ,
                self.__resolution .pdf   )            
            self.__pdf.setBufferFraction ( self.buffer )
            
            if isinstance ( self.bufstrat , int ) and 0 <= self.bufstrat <= 2 : 
                self.__pdf.setBufferStrategy ( self.bufstrat )

            ## set shift-parameters
            if ( not self.shift1 is None ) and ( not self.shift2 is None ) :             
                self.__pdf.setShift ( self.shift1 , self.shift2 )
            
        else :           ##  Use plain numerical integration (could be slow)
            
            assert isinstance ( nsigmas  , num_types ) and 2.5 <= nsigmas , \
                   "Invalid ``nsigmas''  parameter  %s/%s for ``setConvolutionWindow''"  % ( nsigmas , type ( nsigmas ) )
            
            self.__pdf = ROOT.RooNumConvPdf (
                self.old_pdf.new_roo_name ( 'numcnv' ) ,
                'NUM convolution: %s (*) %s' %  ( pdf.name , self.resolution.name ) ,
                self.__xvar            ,
                self.__old_pdf    .pdf ,
                self.__resolution .pdf )
            
            if hasattr ( self.__resolution , 'sigma' ) :                
                if hasattr ( self.__resolution , 'mean' ) :
                    self.__pdf.setConvolutionWindow ( self.__resolution.mean  ,
                                                      self.__resolution.sigma , self.__nsigmas  )
                    logger.debug('Convolution: choose window of %s' % self.__nsigmas ) 
    @property
    def xvar (self ) :
        """The axis variable for  convolution"""
        return self.__xvar
    @property
    def useFFT  ( self ) :
        """``useFFT'' :    use Fast Fourier Transform?"""
        return self.__useFFT
    @property
    def resolution ( self  ) :
        """``resolution'': pdf for resolution function"""
        return self.__resolution 
    @property
    def old_pdf ( self ) :
        """``old'' - the original pdf before convolution"""
        return self.__old_pdf
    @property
    def pdf ( self ) :
        """``new'' (convoluted) PDF"""
        return self.__pdf
    @property
    def nbinsFFT ( self ) :
        """number of cache bins for Fast Fourier Transform"""
        return abs ( self.__nbins )
    @property
    def nbins    ( self ) :
        """number of cache bins for Fast Fourier Transform"""
        return abs ( self.__nbins ) 
    @property
    def buffer ( self ) :
        """``buffer'' : buffer fraction for Fast Fourier Transform"""
        return self.__buffer
    @property
    def bufstrat ( self ) :
        """``bufstrat'' : buffer strategy:
        - 'Extend/0' means is that the input p.d.f convolution observable range is widened to include the buffer range
        - 'Flat/1'   means that the buffer is filled with the p.d.f. value at the boundary of the observable range
        - 'Mirror/2' means that the buffer is filled with a mirror image of the p.d.f. around the convolution observable boundary
        """
        return self.__bufstrat
    @property
    def shift1 ( self ) :
        """''shift1' : parameter for RooFFTConvPdf"""
        return self.__shift1
    @property
    def shift2 ( self ) :
        """''shift2' : parameter for RooFFTConvPdf"""
        return self.__shift2
    
    @property
    def nsigmas ( self ) :
        """``nsigmas'' : convolution window for RooNumConvPdf"""
        return self.__nsigmas
    @property
    def name    ( self ) :
        """``name'' : name of this convoltuoon object/name of pdf"""
        return self.__pdf.name

    def __str__ ( self ) :
        part1 = "pdf=%s,resolution=%s"                      % ( self.old_pdf    ,
                                                                self.resolution ) 
        part2  = "useFFT=%s,nbins=%s,buffer=%s,bufstrat=%s" % ( self.useFFt     , 
                                                                self.nbins      ,
                                                                self.buffer     , 
                                                                self.bufstrat   )  
        return "Convolution(%s,%s,nsigmas=%s)" %  ( part1 , part2 , self.nsigmas)
    __repr__ = __str__
    
# =============================================================================
## @class Convolution_pdf
#  Helper class to simplify the convolutions
#  @code
#  pdf = ...
#  pdfc = Convolution_pdf( pdf  , xvar = ... , resolution = ... , useFFT = True )
#  @endcode
class Convolution_pdf(PDF1) :
    """Helper class/PDF to simplify the convolution:
    >>> pdf = ...
    >>> pdfc = Convolution_pdf( pdf  , xvar = ... , resolution = ... , useFFT = True )
    """
    def __init__ ( self             ,
                   pdf              ,   ## the PDF to be convoluted 
                   resolution       ,   ## the convolution/resolution
                   xvar     = None  ,   ## the axis varable
                   name     = ''    ,   ## name
                   **kwargs         ) : ## convolution configuration 
        
        self.__arg_pdf        = pdf
        self.__arg_resolution = resolution 
        self.__arg_xvar       = xvar
        
        if   isinstance ( pdf , PDF1 ) :
            self.__old_pdf = pdf 
            xvar           = pdf.xvar
        elif isinstance ( pdf , ROOT.RooAbsPdf ) and xvar :
            self.__old_pdf = Generic1D_pdf  ( pdf , xvar )
        else :
            raise TypeError ("Convolution_pdf: invalid pdf/xvar %s/%s"  % ( pdf , xvar ) ) 

        em = self.old_pdf.pdf.extendMode ()
        if   1 == em : self.warning ( "PDF  ``canBeExtended''" )
        elif 2 == em : self.error   ( "PDF ``mustBeExtended''" )

        ## make the actual convolution
        if isinstance ( resolution , Convolution ) :
            assert resolution.xvar is xvar, "Mismatch in ``xvar'': %s vs %s" % ( xvar , resolution.xvar )
            self.__cnv = resolution
        else :
            self.__cnv = Convolution ( pdf        = self.old_pdf.pdf ,
                                       xvar       = xvar             ,
                                       resolution = resolution       , **kwargs ) 

        name = name if name else self.generate_name ( prefix = '(%s)@(%s)' % ( pdf.name , self.resolution.name ) ) 
                            
        ## initialize the base 
        PDF1.__init__ ( self , name , xvar = xvar )


        ## the  actual convoluted PDF 
        self.pdf = self.__cnv.pdf 

        ## save configuration 
        self.config = {
            'name'       : self.name              ,
            'xvar'       : self.xvar              ,
            'pdf'        : self.old_pdf           ,
            'resolution' : self.__arg_resolution  , ## attention! 
            'useFFT'     : self.cnv.useFFT        ,
            'nbins'      : self.cnv.nbins         ,
            'buffer'     : self.cnv.buffer        ,
            'bufstrat'   : self.cnv.bufstrat      ,
            'nsigmas'    : self.cnv.nsigmas       }

    @property
    def convolution ( self ) :
        """``convolution'' : the actual convolution object (same as ``cnv'')"""
        return self.__cnv
    @property
    def cnv         ( self ) :
        """``cnv'' : the actual convolution object (same as ``convolution'')"""
        return self.__cnv
    @property
    def old_pdf ( self ):
        """``old_pdf''  : original (non-convolved) PDF"""
        return self.__old_pdf
    @property
    def original_pdf ( self ):
        """``original_pdf''  : original (non-convolved) PDF"""
        return self.__old_pdf
    
    @property
    def resolution ( self ) :
        """``resolution'' :  the actual resolution function/PDF"""
        return self.cnv.resolution
        
# =============================================================================
if '__main__' == __name__ :
    
    from ostap.utils.docme import docme
    docme ( __name__ , logger = logger )

# =============================================================================
##                                                                      The END 
# =============================================================================

